import typing
import collections
import collections.abc
from typing_extensions import override
from dataclasses import dataclass, field

__all__ = [
    "R",
    "Parameter",
    "TypeResolver",
    "CallableInfo",
    "InjectionTrace",
    "ParameterResult",
    "DependencyConfiguration",
]

R = typing.TypeVar("R")


@dataclass
class TypeResolver:
    """
    Mark that tells ``fundi.scan.scan`` to set ``Parameter.resolve_by_type`` to True.

    This changes logic of ``fundi.resolve.resolve``, so it uses ``Parameter.annotation``
    to find value in scope instead of ``Parameter.name``
    """

    annotation: type


@dataclass
class Parameter:
    name: str
    annotation: typing.Any
    from_: "CallableInfo[typing.Any] | None"
    default: typing.Any = None
    has_default: bool = False
    resolve_by_type: bool = False
    positional_only: bool = False
    keyword_only: bool = False
    positional_varying: bool = False
    keyword_varying: bool = False


@dataclass
class CallableInfo(typing.Generic[R]):
    call: typing.Callable[..., R]
    use_cache: bool
    async_: bool
    context: bool
    generator: bool
    parameters: list[Parameter]
    return_annotation: typing.Any
    configuration: "DependencyConfiguration | None"
    named_parameters: dict[str, Parameter] = field(init=False)
    key: "CacheKey" = field(init=False)

    def __post_init__(self):
        self.named_parameters = {p.name: p for p in self.parameters}
        self.key = CacheKey(self.call)

    @override
    def __hash__(self) -> int:
        return hash(self.key)

    @override
    def __eq__(self, value: object) -> bool:
        return hash(self.key) == hash(value)

    def _build_values(
        self,
        args: tuple[typing.Any, ...],
        kwargs: collections.abc.MutableMapping[str, typing.Any],
        partial: bool = False,
    ) -> dict[str, typing.Any]:
        values: dict[str, typing.Any] = {}

        args_amount = len(args)

        ix = 0
        for parameter in self.parameters:
            name = parameter.name

            if parameter.keyword_varying:
                values[name] = kwargs
                continue

            if name in kwargs:
                values[name] = kwargs.pop(name)
                continue

            if parameter.positional_varying:
                values[name] = args[ix:]
                ix = args_amount
                continue

            if ix < args_amount:
                values[name] = args[ix]
                ix += 1
                continue

            if parameter.has_default:
                values[name] = parameter.default
                continue

            if not partial:
                raise ValueError(f'Argument for parameter "{parameter.name}" not found')

        return values

    def build_values(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> collections.abc.Mapping[str, typing.Any]:
        return self._build_values(args, kwargs)

    def partial_build_values(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> collections.abc.Mapping[str, typing.Any]:
        return self._build_values(args, kwargs, partial=True)

    def build_arguments(
        self, values: collections.abc.Mapping[str, typing.Any]
    ) -> tuple[tuple[typing.Any, ...], dict[str, typing.Any]]:
        positional: tuple[typing.Any, ...] = ()
        keyword: dict[str, typing.Any] = {}

        for parameter in self.parameters:
            name = parameter.name

            if name not in values:
                raise ValueError(f'Value for "{name}" parameter not found')

            value = values[name]

            if parameter.positional_only:
                positional += (value,)
            elif parameter.positional_varying:
                positional += value
            elif parameter.keyword_only:
                keyword[name] = value
            elif parameter.keyword_varying:
                keyword.update(value)
            else:
                positional += (value,)

        return positional, keyword


class CacheKey:
    __slots__: tuple[str, ...] = ("_hash", "_items")

    def __init__(self, *initial_items: collections.abc.Hashable):
        self._hash: int | None = None
        self._items: list[collections.abc.Hashable] = list(initial_items)

    def add(self, *items: collections.abc.Hashable):
        self._items.extend(items)
        self._hash = None

    @override
    def __hash__(self) -> int:
        if self._hash is not None:
            return self._hash

        self._hash = hash(tuple(self._items))

        return self._hash

    @override
    def __eq__(self, value: typing.Hashable) -> bool:
        return self._hash == hash(value)

    @override
    def __repr__(self) -> str:
        return f"#{hash(self)}"


@dataclass
class ParameterResult:
    parameter: Parameter
    value: typing.Any | None
    dependency: CallableInfo[typing.Any] | None
    resolved: bool


@dataclass
class InjectionTrace:
    info: CallableInfo[typing.Any]
    values: collections.abc.Mapping[str, typing.Any]
    origin: "InjectionTrace | None" = None


@dataclass
class DependencyConfiguration:
    configurator: CallableInfo[typing.Any]
    values: collections.abc.Mapping[str, typing.Any]



import typing
import inspect
from dataclasses import replace
from types import BuiltinFunctionType, FunctionType, MethodType
from collections.abc import AsyncGenerator, Awaitable, Generator
from contextlib import AbstractAsyncContextManager, AbstractContextManager

from fundi.util import is_configured, get_configuration, normalize_annotation
from fundi.types import R, CallableInfo, Parameter, TypeResolver


def _transform_parameter(parameter: inspect.Parameter) -> Parameter:
    positional_varying = parameter.kind == inspect.Parameter.VAR_POSITIONAL
    positional_only = parameter.kind == inspect.Parameter.POSITIONAL_ONLY
    keyword_varying = parameter.kind == inspect.Parameter.VAR_KEYWORD
    keyword_only = parameter.kind == inspect.Parameter.KEYWORD_ONLY

    default = parameter.default
    has_default = default is not inspect.Parameter.empty
    from_: CallableInfo[typing.Any] | None = None
    resolve_by_type = False

    if isinstance(default, CallableInfo):
        has_default = False
        from_ = typing.cast(CallableInfo[typing.Any], default)

    annotation = parameter.annotation
    if isinstance(annotation, TypeResolver):
        annotation = annotation.annotation
        resolve_by_type = True

    elif typing.get_origin(annotation) is typing.Annotated and from_ is None:
        args = typing.get_args(annotation)

        if TypeResolver in args:
            resolve_by_type = True
        else:
            presence: tuple[CallableInfo[typing.Any]] | tuple[()] = tuple(
                filter(lambda x: isinstance(x, CallableInfo), args)
            )
            if presence:
                from_ = presence[0]

    return Parameter(
        parameter.name,
        annotation,
        from_=from_,
        default=default if has_default else None,
        has_default=has_default,
        resolve_by_type=resolve_by_type,
        positional_varying=positional_varying,
        positional_only=positional_only,
        keyword_varying=keyword_varying,
        keyword_only=keyword_only,
    )


def _is_context(call: typing.Any):
    if isinstance(call, type):
        return issubclass(call, AbstractContextManager)
    else:
        return isinstance(call, AbstractContextManager)


def _is_async_context(call: typing.Any):
    if isinstance(call, type):
        return issubclass(call, AbstractAsyncContextManager)
    else:
        return isinstance(call, AbstractAsyncContextManager)


def scan(
    call: typing.Callable[..., R],
    caching: bool = True,
    async_: bool | None = None,
    generator: bool | None = None,
    context: bool | None = None,
    use_return_annotation: bool = True,
) -> CallableInfo[R]:
    """
    Get callable information

    :param call: callable to get information from
    :param caching:  whether to use cached result of this callable or not
    :param async_: Override "async_" attriubute value
    :param generator: Override "generator" attriubute value
    :param context: Override "context" attriubute value
    :param use_return_annotation: Whether to use call's return
        annotation to define it's type

    :return: callable information
    """

    if hasattr(call, "__fundi_info__"):
        info = typing.cast(CallableInfo[typing.Any], getattr(call, "__fundi_info__"))

        overrides = {"use_cache": caching}
        if async_ is not None:
            overrides["async_"] = async_

        if generator is not None:
            overrides["generator"] = generator

        if context is not None:
            overrides["context"] = context

        return replace(info, **overrides)

    if not callable(call):
        raise ValueError(
            f"Callable expected, got {type(call)!r}"
        )  # pyright: ignore[reportUnreachable]

    truecall = call.__call__
    if isinstance(call, (FunctionType, BuiltinFunctionType, MethodType, type)):
        truecall = call

    signature = inspect.signature(truecall)

    return_: type[typing.Any] = type
    if signature.return_annotation is not signature.empty:
        annotation = normalize_annotation(signature.return_annotation)[0]

        if not isinstance(annotation, type):
            return_ = type(return_)
        else:
            return_ = annotation

    # WARNING: over-engineered logic!! :3

    _generator: bool = inspect.isgeneratorfunction(truecall)
    _agenerator: bool = inspect.isasyncgenfunction(truecall)
    _context: bool = _is_context(call)
    _acontext: bool = _is_async_context(call)

    # Getting "generator" using return typehint or __code__ flags
    if generator is None:
        generator = (
            use_return_annotation
            and (issubclass(return_, Generator) or issubclass(return_, AsyncGenerator))
        ) or (_generator or _agenerator)

    # Getting "context" using return typehint or callable type
    if context is None:
        context = (
            use_return_annotation
            and (issubclass(return_, (AbstractContextManager, AbstractAsyncContextManager)))
        ) or (_context or _acontext)

    # Getting "async_" using return typehint or __code__ flags or defined above variables
    if async_ is None:
        async_ = (
            use_return_annotation
            and issubclass(return_, (AsyncGenerator, AbstractAsyncContextManager, Awaitable))
        ) or (_agenerator or _acontext or inspect.iscoroutinefunction(truecall))

    parameters = [_transform_parameter(parameter) for parameter in signature.parameters.values()]

    info = CallableInfo(
        call=call,
        use_cache=caching,
        async_=async_,
        context=context,
        generator=generator,
        parameters=parameters,
        return_annotation=signature.return_annotation,
        configuration=get_configuration(call) if is_configured(call) else None,
    )

    try:
        setattr(call, "__fundi_info__", info)
    except (AttributeError, TypeError):
        pass

    return info
